@@ -9,6 +9,10 @@
  *  @{
  */
 
+/*! \brief Access SCTP flags from the msgb control buffer */
+#define OSMO_STREAM_SCTP_MSG_FLAGS_NOTIFICATION 0x80 /* sctp_recvmsg() flags=MSG_NOTIFICATION, msgb_data() contains "union sctp_notification*" */
+#define msgb_sctp_msg_flags(msg)	(msg)->cb[2]
+
 /*! \brief Access the SCTP PPID from the msgb control buffer */
 #define msgb_sctp_ppid(msg)             (msg)->cb[3]
 /*! \brief Access the SCTP Stream ID from the msgb control buffer */
@@ -25,6 +29,8 @@ void osmo_stream_srv_link_set_addr(struct osmo_stream_srv_link *link, const char
 int osmo_stream_srv_link_set_addrs(struct osmo_stream_srv_link *link, const char **addr, size_t addrcnt);
 void osmo_stream_srv_link_set_port(struct osmo_stream_srv_link *link, uint16_t port);
 void osmo_stream_srv_link_set_proto(struct osmo_stream_srv_link *link, uint16_t proto);
+int osmo_stream_srv_link_set_type(struct osmo_stream_srv_link *link, int type);
+int osmo_stream_srv_link_set_domain(struct osmo_stream_srv_link *link, int domain);
 void osmo_stream_srv_link_set_accept_cb(struct osmo_stream_srv_link *link, int (*accept_cb)(struct osmo_stream_srv_link *link, int fd));
 void osmo_stream_srv_link_set_data(struct osmo_stream_srv_link *link, void *data);
 void *osmo_stream_srv_link_get_data(struct osmo_stream_srv_link *link);
@@ -50,6 +56,8 @@ void osmo_stream_srv_set_data(struct osmo_stream_srv *conn, void *data);
 void osmo_stream_srv_send(struct osmo_stream_srv *conn, struct msgb *msg);
 int osmo_stream_srv_recv(struct osmo_stream_srv *conn, struct msgb *msg);
 
+void osmo_stream_srv_clear_tx_queue(struct osmo_stream_srv *conn);
+
 /*! \brief Osmocom Stream Client: Single client connection */
 struct osmo_stream_cli;
 
@@ -57,6 +65,8 @@ void osmo_stream_cli_set_nodelay(struct osmo_stream_cli *cli, bool nodelay);
 void osmo_stream_cli_set_addr(struct osmo_stream_cli *cli, const char *addr);
 int osmo_stream_cli_set_addrs(struct osmo_stream_cli *cli, const char **addr, size_t addrcnt);
 void osmo_stream_cli_set_port(struct osmo_stream_cli *cli, uint16_t port);
+int osmo_stream_cli_set_type(struct osmo_stream_cli *cli, int type);
+int osmo_stream_cli_set_domain(struct osmo_stream_cli *cli, int domain);
 void osmo_stream_cli_set_proto(struct osmo_stream_cli *cli, uint16_t proto);
 void osmo_stream_cli_set_local_addr(struct osmo_stream_cli *cli, const char *addr);
 int osmo_stream_cli_set_local_addrs(struct osmo_stream_cli *cli, const char **addr, size_t addrcnt);
@@ -81,6 +91,8 @@ int osmo_stream_cli_open2(struct osmo_stream_cli *cli, int reconnect) \
 void osmo_stream_cli_close(struct osmo_stream_cli *cli);
 
 void osmo_stream_cli_send(struct osmo_stream_cli *cli, struct msgb *msg);
-int osmo_stream_cli_recv(struct osmo_stream_cli *conn, struct msgb *msg);
+int osmo_stream_cli_recv(struct osmo_stream_cli *cli, struct msgb *msg);
+
+void osmo_stream_cli_clear_tx_queue(struct osmo_stream_cli *cli);
 
 /*! @} */
