@@ -49,6 +49,8 @@
 #include <netinet/sctp.h>
 #endif
 
+#include <osmocom/netif/sctp.h>
+
 #define LOGSCLI(cli, level, fmt, args...) \
 	LOGP(DLINP, level, "[%s] %s(): " fmt, get_value_string(stream_cli_state_names, (cli)->state), __func__, ## args)
 
@@ -87,7 +89,7 @@ static int byte_nonzero(const uint8_t *u8, unsigned int offset, unsigned int u8_
 	return -1;
 }
 
-static int sctp_sockopt_event_subscribe_size = 0;
+static unsigned int sctp_sockopt_event_subscribe_size = 0;
 
 static int determine_sctp_sockopt_event_subscribe_size(void)
 {
@@ -96,7 +98,7 @@ static int determine_sctp_sockopt_event_subscribe_size(void)
 	int sd, rc;
 
 	/* only do this once */
-	if (sctp_sockopt_event_subscribe_size != 0)
+	if (sctp_sockopt_event_subscribe_size > 0)
 		return 0;
 
 	sd = socket(AF_INET, SOCK_STREAM, IPPROTO_SCTP);
@@ -108,7 +110,7 @@ static int determine_sctp_sockopt_event_subscribe_size(void)
 	if (rc < 0)
 		return rc;
 
-	sctp_sockopt_event_subscribe_size = buf_len;
+	sctp_sockopt_event_subscribe_size = (unsigned int)buf_len;
 
 	LOGP(DLINP, LOGL_INFO, "sizes of 'struct sctp_event_subscribe': compile-time %zu, kernel: %u\n",
 		sizeof(struct sctp_event_subscribe), sctp_sockopt_event_subscribe_size);
@@ -256,6 +258,8 @@ struct osmo_stream_cli {
 	char				*local_addr[OSMO_STREAM_MAX_ADDRS];
 	uint8_t			 	local_addrcnt;
 	uint16_t			local_port;
+	int				sk_domain;
+	int				sk_type;
 	uint16_t			proto;
 	int (*connect_cb)(struct osmo_stream_cli *srv);
 	int (*disconnect_cb)(struct osmo_stream_cli *srv);
@@ -331,16 +335,14 @@ static int osmo_stream_cli_write(struct osmo_stream_cli *cli)
 	struct sctp_sndrcvinfo sinfo;
 #endif
 	struct msgb *msg;
-	struct llist_head *lh;
 	int ret;
 
 	if (llist_empty(&cli->tx_queue)) {
 		osmo_fd_write_disable(&cli->ofd);
 		return 0;
 	}
-	lh = cli->tx_queue.next;
-	llist_del(lh);
-	msg = llist_entry(lh, struct msgb, list);
+	msg = llist_first_entry(&cli->tx_queue, struct msgb, list);
+	llist_del(&msg->list);
 
 	if (!osmo_stream_cli_is_connected(cli)) {
 		LOGSCLI(cli, LOGL_ERROR, "not connected, dropping data!\n");
@@ -349,31 +351,70 @@ static int osmo_stream_cli_write(struct osmo_stream_cli *cli)
 
 	LOGSCLI(cli, LOGL_DEBUG, "sending %u bytes of data\n", msgb_length(msg));
 
-	switch (cli->proto) {
+	switch (cli->sk_domain) {
+	case AF_UNIX:
+		ret = send(cli->ofd.fd, msgb_data(msg), msgb_length(msg), 0);
+		break;
+	case AF_UNSPEC:
+	case AF_INET:
+	case AF_INET6:
+		switch (cli->proto) {
 #ifdef HAVE_LIBSCTP
-	case IPPROTO_SCTP:
-		memset(&sinfo, 0, sizeof(sinfo));
-		sinfo.sinfo_ppid = htonl(msgb_sctp_ppid(msg));
-		sinfo.sinfo_stream = msgb_sctp_stream(msg);
-		ret = sctp_send(cli->ofd.fd, msg->data, msgb_length(msg),
-				&sinfo, MSG_NOSIGNAL);
-		break;
+		case IPPROTO_SCTP:
+			memset(&sinfo, 0, sizeof(sinfo));
+			sinfo.sinfo_ppid = htonl(msgb_sctp_ppid(msg));
+			sinfo.sinfo_stream = msgb_sctp_stream(msg);
+			ret = sctp_send(cli->ofd.fd, msgb_data(msg), msgb_length(msg),
+					&sinfo, MSG_NOSIGNAL);
+			break;
 #endif
-	case IPPROTO_TCP:
-	default:
-		ret = send(cli->ofd.fd, msg->data, msgb_length(msg), 0);
+		case IPPROTO_TCP:
+		default:
+			ret = send(cli->ofd.fd, msgb_data(msg), msgb_length(msg), 0);
+			break;
+		}
 		break;
+	default:
+		ret = -ENOTSUP;
 	}
+
+	if (ret >= 0 && ret < msgb_length(msg)) {
+		LOGP(DLINP, LOGL_ERROR, "short send: %d < exp %u\n", ret, msgb_length(msg));
+		/* Update msgb and re-add it at the start of the queue: */
+		msgb_pull(msg, ret);
+		llist_add(&msg->list, &cli->tx_queue);
+		return 0;
+	}
+
 	if (ret < 0) {
 		if (errno == EPIPE || errno == ENOTCONN) {
 			osmo_stream_cli_reconnect(cli);
 		}
 		LOGSCLI(cli, LOGL_ERROR, "error %d to send\n", ret);
 	}
+
 	msgb_free(msg);
+
+	if (llist_empty(&cli->tx_queue))
+		osmo_fd_write_disable(&cli->ofd);
+
 	return 0;
 }
 
+static int _setsockopt_nosigpipe(struct osmo_stream_cli *cli)
+{
+#ifdef SO_NOSIGPIPE
+	int ret;
+	int val = 1;
+	ret = setsockopt(cli->ofd.fd, SOL_SOCKET, SO_NOSIGPIPE, (void *)&val, sizeof(val));
+	if (ret < 0)
+		LOGSCLI(cli, LOGL_DEBUG, "Failed setting SO_NOSIGPIPE: %s\n", strerror(errno));
+	return ret;
+#else
+	return 0;
+#endif
+}
+
 static int osmo_stream_cli_fd_cb(struct osmo_fd *ofd, unsigned int what)
 {
 	struct osmo_stream_cli *cli = ofd->data;
@@ -395,15 +436,20 @@ static int osmo_stream_cli_fd_cb(struct osmo_fd *ofd, unsigned int what)
 
 		LOGSCLI(cli, LOGL_DEBUG, "connection done.\n");
 		cli->state = STREAM_CLI_STATE_CONNECTED;
-		if (cli->proto == IPPROTO_SCTP) {
-#ifdef SO_NOSIGPIPE
-			int val = 1;
-
-			ret = setsockopt(ofd->fd, SOL_SOCKET, SO_NOSIGPIPE, (void*)&val, sizeof(val));
-			if (ret < 0)
-				LOGSCLI(cli, LOGL_DEBUG, "Failed setting SO_NOSIGPIPE: %s\n", strerror(errno));
-#endif
-			sctp_sock_activate_events(ofd->fd);
+		switch (cli->sk_domain) {
+		case AF_UNIX:
+			_setsockopt_nosigpipe(cli);
+			break;
+		case AF_UNSPEC:
+		case AF_INET:
+		case AF_INET6:
+			if (cli->proto == IPPROTO_SCTP) {
+				_setsockopt_nosigpipe(cli);
+				sctp_sock_activate_events(ofd->fd);
+			}
+			break;
+		default:
+			break;
 		}
 		if (cli->connect_cb)
 			cli->connect_cb(cli);
@@ -441,6 +487,8 @@ struct osmo_stream_cli *osmo_stream_cli_create(void *ctx)
 	if (!cli)
 		return NULL;
 
+	cli->sk_domain = AF_UNSPEC;
+	cli->sk_type = SOCK_STREAM;
 	cli->proto = IPPROTO_TCP;
 	cli->ofd.fd = -1;
 	cli->ofd.priv_nr = 0;	/* XXX */
@@ -557,6 +605,46 @@ osmo_stream_cli_set_proto(struct osmo_stream_cli *cli, uint16_t proto)
 	cli->flags |= OSMO_STREAM_CLI_F_RECONF;
 }
 
+/*! \brief Set the socket type for the stream server link
+ *  \param[in] cli Stream Client to modify
+ *  \param[in] type Socket Type (like SOCK_STREAM (default), SOCK_SEQPACKET, ...)
+ *  \returns zero on success, negative on error.
+ */
+int osmo_stream_cli_set_type(struct osmo_stream_cli *cli, int type)
+{
+	switch (type) {
+	case SOCK_STREAM:
+	case SOCK_SEQPACKET:
+		break;
+	default:
+		return -ENOTSUP;
+	}
+	cli->sk_type = type;
+	cli->flags |= OSMO_STREAM_CLI_F_RECONF;
+	return 0;
+}
+
+/*! \brief Set the socket type for the stream server link
+ *  \param[in] cli Stream Client to modify
+ *  \param[in] type Socket Domain (like AF_UNSPEC (default for IP), AF_UNIX, AF_INET, ...)
+ *  \returns zero on success, negative on error.
+ */
+int osmo_stream_cli_set_domain(struct osmo_stream_cli *cli, int domain)
+{
+	switch (domain) {
+	case AF_UNSPEC:
+	case AF_INET:
+	case AF_INET6:
+	case AF_UNIX:
+		break;
+	default:
+		return -ENOTSUP;
+	}
+	cli->sk_domain = domain;
+	cli->flags |= OSMO_STREAM_CLI_F_RECONF;
+	return 0;
+}
+
 /*! \brief Set the reconnect time of the stream client socket
  *  \param[in] cli Stream Client to modify
  *  \param[in] timeout Re-connect timeout in seconds or negative value to disable auto-reconnection */
@@ -731,21 +819,31 @@ int osmo_stream_cli_open(struct osmo_stream_cli *cli)
 
 	cli->flags &= ~OSMO_STREAM_CLI_F_RECONF;
 
-
-	switch (cli->proto) {
-#ifdef HAVE_LIBSCTP
-	case IPPROTO_SCTP:
-		ret = osmo_sock_init2_multiaddr(AF_UNSPEC, SOCK_STREAM, cli->proto,
-						(const char **)cli->local_addr, cli->local_addrcnt, cli->local_port,
-						(const char **)cli->addr, cli->addrcnt, cli->port,
-						OSMO_SOCK_F_CONNECT|OSMO_SOCK_F_BIND|OSMO_SOCK_F_NONBLOCK);
+	switch (cli->sk_domain) {
+	case AF_UNIX:
+		ret = osmo_sock_unix_init(cli->sk_type, 0, cli->addr[0], OSMO_SOCK_F_CONNECT|OSMO_SOCK_F_BIND|OSMO_SOCK_F_NONBLOCK);
 		break;
+	case AF_INET:
+	case AF_INET6:
+	case AF_UNSPEC:
+		switch (cli->proto) {
+#ifdef HAVE_LIBSCTP
+		case IPPROTO_SCTP:
+			ret = osmo_sock_init2_multiaddr(cli->sk_domain, cli->sk_type, cli->proto,
+							(const char **)cli->local_addr, cli->local_addrcnt, cli->local_port,
+							(const char **)cli->addr, cli->addrcnt, cli->port,
+							OSMO_SOCK_F_CONNECT|OSMO_SOCK_F_BIND|OSMO_SOCK_F_NONBLOCK);
+			break;
 #endif
+		default:
+			ret = osmo_sock_init2(cli->sk_domain, cli->sk_type, cli->proto,
+					      cli->local_addr[0], cli->local_port,
+					      cli->addr[0], cli->port,
+					      OSMO_SOCK_F_CONNECT|OSMO_SOCK_F_BIND|OSMO_SOCK_F_NONBLOCK);
+		}
+		break;
 	default:
-		ret = osmo_sock_init2(AF_UNSPEC, SOCK_STREAM, cli->proto,
-				      cli->local_addr[0], cli->local_port,
-				      cli->addr[0], cli->port,
-				      OSMO_SOCK_F_CONNECT|OSMO_SOCK_F_BIND|OSMO_SOCK_F_NONBLOCK);
+		return -ENOTSUP;
 	}
 
 	if (ret < 0) {
@@ -786,6 +884,8 @@ static void cli_timer_cb(void *data)
  *  \param[in] msg Message buffer to enqueue in transmit queue */
 void osmo_stream_cli_send(struct osmo_stream_cli *cli, struct msgb *msg)
 {
+	OSMO_ASSERT(cli);
+	OSMO_ASSERT(msg);
 	msgb_enqueue(&cli->tx_queue, msg);
 	osmo_fd_write_enable(&cli->ofd);
 }
@@ -797,6 +897,8 @@ void osmo_stream_cli_send(struct osmo_stream_cli *cli, struct msgb *msg)
 int osmo_stream_cli_recv(struct osmo_stream_cli *cli, struct msgb *msg)
 {
 	int ret;
+	OSMO_ASSERT(cli);
+	OSMO_ASSERT(msg);
 
 	ret = recv(cli->ofd.fd, msg->data, msg->data_len, 0);
 	if (ret < 0) {
@@ -815,6 +917,15 @@ int osmo_stream_cli_recv(struct osmo_stream_cli *cli, struct msgb *msg)
 	return ret;
 }
 
+void osmo_stream_cli_clear_tx_queue(struct osmo_stream_cli *cli)
+{
+	msgb_queue_free(&cli->tx_queue);
+	/* If in state 'connecting', keep WRITE flag up to receive
+	 * socket connection signal and then transition to STATE_CONNECTED: */
+	if (cli->state == STREAM_CLI_STATE_CONNECTED)
+		osmo_fd_write_disable(&cli->ofd);
+}
+
 /*
  * Server side.
  */
@@ -827,6 +938,8 @@ struct osmo_stream_srv_link {
 	char			*addr[OSMO_STREAM_MAX_ADDRS];
 	uint8_t			addrcnt;
 	uint16_t		port;
+	int			sk_domain;
+	int			sk_type;
 	uint16_t		proto;
 	int (*accept_cb)(struct osmo_stream_srv_link *srv, int fd);
 	void			*data;
@@ -838,7 +951,7 @@ static int osmo_stream_srv_fd_cb(struct osmo_fd *ofd, unsigned int what)
 	int ret;
 	int sock_fd;
 	char addrstr[128];
-	bool is_ipv6;
+	bool is_ipv6 = false;
 	struct sockaddr_storage sa;
 	socklen_t sa_len = sizeof(sa);
 	struct osmo_stream_srv_link *link = ofd->data;
@@ -849,19 +962,35 @@ static int osmo_stream_srv_fd_cb(struct osmo_fd *ofd, unsigned int what)
 			"peer, reason=`%s'\n", strerror(errno));
 		return ret;
 	}
-	is_ipv6 = ((struct sockaddr *)&sa)->sa_family == AF_INET6;
-	LOGP(DLINP, LOGL_DEBUG, "accept()ed new link from %s to port %u\n",
-		inet_ntop(is_ipv6 ? AF_INET6 : AF_INET,
-			  is_ipv6 ? (void*)&(((struct sockaddr_in6 *)&sa)->sin6_addr) :
-				    (void*)&(((struct sockaddr_in *)&sa)->sin_addr),
-			  addrstr, sizeof(addrstr)),
-		link->port);
 	sock_fd = ret;
 
-	if (link->proto == IPPROTO_SCTP) {
-		ret = sctp_sock_activate_events(sock_fd);
-		if (ret < 0)
-			goto error_close_socket;
+	is_ipv6 = false;
+	switch (((struct sockaddr *)&sa)->sa_family) {
+	case AF_UNIX:
+		LOGP(DLINP, LOGL_DEBUG, "accept()ed new link on fd %d\n",
+		     sock_fd);
+		break;
+	case AF_INET6:
+		is_ipv6 = true;
+		/* fall through */
+	case AF_INET:
+		LOGP(DLINP, LOGL_DEBUG, "accept()ed new link from %s to port %u\n",
+			inet_ntop(is_ipv6 ? AF_INET6 : AF_INET,
+				  is_ipv6 ? (void *)&(((struct sockaddr_in6 *)&sa)->sin6_addr) :
+					    (void *)&(((struct sockaddr_in *)&sa)->sin_addr),
+				  addrstr, sizeof(addrstr)),
+			link->port);
+
+		if (link->proto == IPPROTO_SCTP) {
+			ret = sctp_sock_activate_events(sock_fd);
+			if (ret < 0)
+				goto error_close_socket;
+		}
+		break;
+	default:
+		LOGP(DLINP, LOGL_DEBUG, "accept()ed unexpected address family %d\n",
+		     ((struct sockaddr *)&sa)->sa_family);
+		goto error_close_socket;
 	}
 
 	if (link->flags & OSMO_STREAM_SRV_F_NODELAY) {
@@ -899,6 +1028,8 @@ struct osmo_stream_srv_link *osmo_stream_srv_link_create(void *ctx)
 	if (!link)
 		return NULL;
 
+	link->sk_domain = AF_UNSPEC;
+	link->sk_type = SOCK_STREAM;
 	link->proto = IPPROTO_TCP;
 	osmo_fd_setup(&link->ofd, -1, OSMO_FD_READ | OSMO_FD_WRITE, osmo_stream_srv_fd_cb, link, 0);
 
@@ -978,6 +1109,47 @@ osmo_stream_srv_link_set_proto(struct osmo_stream_srv_link *link,
 	link->flags |= OSMO_STREAM_SRV_F_RECONF;
 }
 
+
+/*! \brief Set the socket type for the stream server link
+ *  \param[in] link Stream Server Link to modify
+ *  \param[in] type Socket Type (like SOCK_STREAM (default), SOCK_SEQPACKET, ...)
+ *  \returns zero on success, negative on error.
+ */
+int osmo_stream_srv_link_set_type(struct osmo_stream_srv_link *link, int type)
+{
+	switch (type) {
+	case SOCK_STREAM:
+	case SOCK_SEQPACKET:
+		break;
+	default:
+		return -ENOTSUP;
+	}
+	link->sk_type = type;
+	link->flags |= OSMO_STREAM_SRV_F_RECONF;
+	return 0;
+}
+
+/*! \brief Set the socket type for the stream server link
+ *  \param[in] link Stream Server Link to modify
+ *  \param[in] type Socket Domain (like AF_UNSPEC (default for IP), AF_UNIX, AF_INET, ...)
+ *  \returns zero on success, negative on error.
+ */
+int osmo_stream_srv_link_set_domain(struct osmo_stream_srv_link *link, int domain)
+{
+	switch (domain) {
+	case AF_UNSPEC:
+	case AF_INET:
+	case AF_INET6:
+	case AF_UNIX:
+		break;
+	default:
+		return -ENOTSUP;
+	}
+	link->sk_domain = domain;
+	link->flags |= OSMO_STREAM_SRV_F_RECONF;
+	return 0;
+}
+
 /*! \brief Set application private data of the stream server link
  *  \param[in] link Stream Server Link to modify
  *  \param[in] data User-specific data (available in call-back functions) */
@@ -1060,17 +1232,28 @@ int osmo_stream_srv_link_open(struct osmo_stream_srv_link *link)
 
 	link->flags &= ~OSMO_STREAM_SRV_F_RECONF;
 
-	switch (link->proto) {
-#ifdef HAVE_LIBSCTP
-	case IPPROTO_SCTP:
-		ret = osmo_sock_init2_multiaddr(AF_UNSPEC, SOCK_STREAM, link->proto,
-						(const char **)link->addr, link->addrcnt, link->port,
-						NULL, 0, 0, OSMO_SOCK_F_BIND);
+	switch (link->sk_domain) {
+	case AF_UNIX:
+		ret = osmo_sock_unix_init(link->sk_type, 0, link->addr[0], OSMO_SOCK_F_BIND);
 		break;
+	case AF_UNSPEC:
+	case AF_INET:
+	case AF_INET6:
+		switch (link->proto) {
+#ifdef HAVE_LIBSCTP
+		case IPPROTO_SCTP:
+			ret = osmo_sock_init2_multiaddr(link->sk_domain, link->sk_type, link->proto,
+							(const char **)link->addr, link->addrcnt, link->port,
+							NULL, 0, 0, OSMO_SOCK_F_BIND);
+			break;
 #endif
+		default:
+			ret = osmo_sock_init(link->sk_domain, link->sk_type, link->proto,
+					     link->addr[0], link->port, OSMO_SOCK_F_BIND);
+		}
+		break;
 	default:
-		ret = osmo_sock_init(AF_UNSPEC, SOCK_STREAM, link->proto,
-					link->addr[0], link->port, OSMO_SOCK_F_BIND);
+		ret = -ENOTSUP;
 	}
 	if (ret < 0)
 		return ret;
@@ -1131,41 +1314,63 @@ static void osmo_stream_srv_write(struct osmo_stream_srv *conn)
 	struct sctp_sndrcvinfo sinfo;
 #endif
 	struct msgb *msg;
-	struct llist_head *lh;
 	int ret;
 
-	LOGP(DLINP, LOGL_DEBUG, "sending data\n");
-
 	if (llist_empty(&conn->tx_queue)) {
 		osmo_fd_write_disable(&conn->ofd);
 		return;
 	}
-	lh = conn->tx_queue.next;
-	llist_del(lh);
-	msg = llist_entry(lh, struct msgb, list);
+	msg = llist_first_entry(&conn->tx_queue, struct msgb, list);
+	llist_del(&msg->list);
 
-	switch (conn->srv->proto) {
+	LOGP(DLINP, LOGL_DEBUG, "sending %u bytes of data\n", msg->len);
+
+	switch (conn->srv->sk_domain) {
+	case AF_UNIX:
+		ret = send(conn->ofd.fd, msgb_data(msg), msgb_length(msg), 0);
+		break;
+	case AF_INET:
+	case AF_INET6:
+	case AF_UNSPEC:
+		switch (conn->srv->proto) {
 #ifdef HAVE_LIBSCTP
-	case IPPROTO_SCTP:
-		memset(&sinfo, 0, sizeof(sinfo));
-		sinfo.sinfo_ppid = htonl(msgb_sctp_ppid(msg));
-		sinfo.sinfo_stream = msgb_sctp_stream(msg);
-		ret = sctp_send(conn->ofd.fd, msg->data, msgb_length(msg),
-				&sinfo, MSG_NOSIGNAL);
-		break;
+		case IPPROTO_SCTP:
+			memset(&sinfo, 0, sizeof(sinfo));
+			sinfo.sinfo_ppid = htonl(msgb_sctp_ppid(msg));
+			sinfo.sinfo_stream = msgb_sctp_stream(msg);
+			ret = sctp_send(conn->ofd.fd, msgb_data(msg), msgb_length(msg),
+					&sinfo, MSG_NOSIGNAL);
+			break;
 #endif
-	case IPPROTO_TCP:
-	default:
-		ret = send(conn->ofd.fd, msg->data, msg->len, 0);
+		case IPPROTO_TCP:
+		default:
+			ret = send(conn->ofd.fd, msgb_data(msg), msgb_length(msg), 0);
+			break;
+		}
 		break;
+	default:
+		ret = -1;
+		errno = ENOTSUP;
 	}
-	if (ret < 0) {
-		LOGP(DLINP, LOGL_ERROR, "error to send\n");
+
+	if (ret >= 0 && ret < msgb_length(msg)) {
+		LOGP(DLINP, LOGL_ERROR, "short send: %d < exp %u\n", ret, msgb_length(msg));
+		/* Update msgb and re-add it at the start of the queue: */
+		msgb_pull(msg, ret);
+		llist_add(&msg->list, &conn->tx_queue);
+		return;
 	}
+
+	if (ret == -1) /* send(): On error -1 is returned, and errno is set appropriately */
+		LOGP(DLINP, LOGL_ERROR, "error to send: %s\n", strerror(errno));
+
 	msgb_free(msg);
 
-	if (llist_empty(&conn->tx_queue) && (conn->flags & OSMO_STREAM_SRV_F_FLUSH_DESTROY))
-		osmo_stream_srv_destroy(conn);
+	if (llist_empty(&conn->tx_queue)) {
+		osmo_fd_write_disable(&conn->ofd);
+		if (conn->flags & OSMO_STREAM_SRV_F_FLUSH_DESTROY)
+			osmo_stream_srv_destroy(conn);
+	}
 }
 
 static int osmo_stream_srv_cb(struct osmo_fd *ofd, unsigned int what)
@@ -1173,7 +1378,7 @@ static int osmo_stream_srv_cb(struct osmo_fd *ofd, unsigned int what)
 	struct osmo_stream_srv *conn = ofd->data;
 	int rc = 0;
 
-	LOGP(DLINP, LOGL_DEBUG, "connected read/write\n");
+	LOGP(DLINP, LOGL_DEBUG, "connected read/write (what=0x%x)\n", what);
 	if (what & OSMO_FD_READ)
 		rc = osmo_stream_srv_read(conn);
 	if (rc != -EBADF && (what & OSMO_FD_WRITE))
@@ -1194,6 +1399,8 @@ osmo_stream_srv_create(void *ctx, struct osmo_stream_srv_link *link,
 {
 	struct osmo_stream_srv *conn;
 
+	OSMO_ASSERT(link);
+
 	conn = talloc_zero(ctx, struct osmo_stream_srv);
 	if (conn == NULL) {
 		LOGP(DLINP, LOGL_ERROR, "cannot allocate new peer in srv, "
@@ -1268,8 +1475,9 @@ struct osmo_stream_srv_link *osmo_stream_srv_get_master(struct osmo_stream_srv *
  *  \param[in] conn Stream Server to be destroyed */
 void osmo_stream_srv_destroy(struct osmo_stream_srv *conn)
 {
-	close(conn->ofd.fd);
 	osmo_fd_unregister(&conn->ofd);
+	close(conn->ofd.fd);
+	conn->ofd.fd = -1;
 	if (conn->closed_cb)
 		conn->closed_cb(conn);
 	msgb_queue_free(&conn->tx_queue);
@@ -1281,6 +1489,8 @@ void osmo_stream_srv_destroy(struct osmo_stream_srv *conn)
  *  \param[in] msg Message buffer to enqueue in transmit queue */
 void osmo_stream_srv_send(struct osmo_stream_srv *conn, struct msgb *msg)
 {
+	OSMO_ASSERT(conn);
+	OSMO_ASSERT(msg);
 	if (conn->flags & OSMO_STREAM_SRV_F_FLUSH_DESTROY) {
 		LOGP(DLINP, LOGL_DEBUG, "Connection is being flushed and closed; ignoring new outgoing message\n");
 		return;
@@ -1290,80 +1500,123 @@ void osmo_stream_srv_send(struct osmo_stream_srv *conn, struct msgb *msg)
 	osmo_fd_write_enable(&conn->ofd);
 }
 
+#ifdef HAVE_LIBSCTP
+static int _sctp_recvmsg_wrapper(int fd, struct msgb *msg)
+{
+	struct sctp_sndrcvinfo sinfo;
+	int flags = 0;
+	int ret;
+
+	ret = sctp_recvmsg(fd, msgb_data(msg), msgb_tailroom(msg),
+			NULL, NULL, &sinfo, &flags);
+	msgb_sctp_msg_flags(msg) = 0;
+	msgb_sctp_ppid(msg) = ntohl(sinfo.sinfo_ppid);
+	msgb_sctp_stream(msg) = sinfo.sinfo_stream;
+	if (flags & MSG_NOTIFICATION) {
+		union sctp_notification *notif = (union sctp_notification *)msgb_data(msg);
+		LOGP(DLINP, LOGL_DEBUG, "NOTIFICATION %u flags=0x%x\n", notif->sn_header.sn_type, notif->sn_header.sn_flags);
+		msgb_put(msg, sizeof(union sctp_notification));
+		msgb_sctp_msg_flags(msg) = OSMO_STREAM_SCTP_MSG_FLAGS_NOTIFICATION;
+		switch (notif->sn_header.sn_type) {
+		case SCTP_ASSOC_CHANGE:
+			LOGP(DLINP, LOGL_DEBUG, "===> ASSOC CHANGE:");
+			switch (notif->sn_assoc_change.sac_state) {
+			case SCTP_COMM_UP:
+				LOGPC(DLINP, LOGL_DEBUG, " UP\n");
+				break;
+			case SCTP_COMM_LOST:
+				LOGPC(DLINP, LOGL_DEBUG, " LOST\n");
+				/* Handle this like a regular disconnect */
+				return 0;
+			case SCTP_RESTART:
+				LOGPC(DLINP, LOGL_DEBUG, " RESTART\n");
+				break;
+			case SCTP_SHUTDOWN_COMP:
+				LOGPC(DLINP, LOGL_DEBUG, " SHUTDOWN COMP\n");
+				break;
+			case SCTP_CANT_STR_ASSOC:
+				LOGPC(DLINP, LOGL_DEBUG, " CANT STR ASSOC\n");
+				break;
+			}
+			break;
+		case SCTP_SEND_FAILED:
+			LOGP(DLINP, LOGL_DEBUG, "===> SEND FAILED\n");
+			break;
+		case SCTP_PEER_ADDR_CHANGE:
+			{
+			char addr_str[INET6_ADDRSTRLEN + 10];
+			struct sockaddr_storage sa = notif->sn_paddr_change.spc_aaddr;
+			osmo_sockaddr_to_str_buf(addr_str, sizeof(addr_str),
+						 (const struct osmo_sockaddr *)&sa);
+			LOGP(DLINP, LOGL_DEBUG, "===> PEER ADDR CHANGE: %s %s err=%s\n",
+			     addr_str, osmo_sctp_paddr_chg_str(notif->sn_paddr_change.spc_state),
+			     (notif->sn_paddr_change.spc_state == SCTP_ADDR_UNREACHABLE) ?
+				osmo_sctp_sn_error_str(notif->sn_paddr_change.spc_error) : "None");
+			}
+			break;
+		case SCTP_SHUTDOWN_EVENT:
+			LOGP(DLINP, LOGL_DEBUG, "===> SHUTDOWN EVT\n");
+			/* Handle this like a regular disconnect */
+			return 0;
+		}
+		return -EAGAIN;
+	}
+	return ret;
+}
+#endif
+
 /*! \brief Receive data via Osmocom stream server
  *  \param[in] conn Stream Server from which to receive
  *  \param msg pre-allocate message buffer to which received data is appended
  *  \returns number of bytes read, negative on error.
+ *
+ *  If conn is an SCTP connection, additional specific considerations shall be taken:
+ *  - msg->cb is always filled with SCTP ppid, and SCTP stream values, see msgb_sctp_*() APIs.
+ *  - If an SCTP notification was received when reading from the SCTP socket,
+ *    msgb_sctp_msg_flags(msg) will contain bit flag
+ *    OSMO_STREAM_SCTP_MSG_FLAGS_NOTIFICATION set, and the msgb will
+ *    contain a "union sctp_notification" instead of user data. In this case the
+ *    return code will be either 0 (if conn is considered dead after the
+ *    notification) or -EAGAIN (if conn is considered still alive after the
+ *    notification) resembling the standard recv() API.
  */
 int osmo_stream_srv_recv(struct osmo_stream_srv *conn, struct msgb *msg)
 {
-#ifdef HAVE_LIBSCTP
-	struct sctp_sndrcvinfo sinfo;
-	int flags = 0;
-#endif
 	int ret;
+	OSMO_ASSERT(conn);
+	OSMO_ASSERT(msg);
 
-	if (!msg)
-		return -EINVAL;
-
-	switch (conn->srv->proto) {
-#ifdef HAVE_LIBSCTP
-	case IPPROTO_SCTP:
-		ret = sctp_recvmsg(conn->ofd.fd, msgb_data(msg), msgb_tailroom(msg),
-				NULL, NULL, &sinfo, &flags);
-		if (flags & MSG_NOTIFICATION) {
-			union sctp_notification *notif = (union sctp_notification *) msgb_data(msg);
-			LOGP(DLINP, LOGL_DEBUG, "NOTIFICATION %u flags=0x%x\n", notif->sn_header.sn_type, notif->sn_header.sn_flags);
-			switch (notif->sn_header.sn_type) {
-			case SCTP_ASSOC_CHANGE:
-				LOGP(DLINP, LOGL_DEBUG, "===> ASSOC CHANGE:");
-				switch (notif->sn_assoc_change.sac_state) {
-				case SCTP_COMM_UP:
-					LOGPC(DLINP, LOGL_DEBUG, " UP\n");
-					break;
-				case SCTP_COMM_LOST:
-					LOGPC(DLINP, LOGL_DEBUG, " LOST\n");
-					break;
-				case SCTP_RESTART:
-					LOGPC(DLINP, LOGL_DEBUG, " RESTART\n");
-					break;
-				case SCTP_SHUTDOWN_COMP:
-					LOGPC(DLINP, LOGL_DEBUG, " SHUTDOWN COMP\n");
-					break;
-				case SCTP_CANT_STR_ASSOC:
-					LOGPC(DLINP, LOGL_DEBUG, " CANT STR ASSOC\n");
-					break;
-				}
-				break;
-			case SCTP_PEER_ADDR_CHANGE:
-				LOGP(DLINP, LOGL_DEBUG, "===> PEER ADDR CHANGE\n");
-				break;
-			case SCTP_SHUTDOWN_EVENT:
-				LOGP(DLINP, LOGL_DEBUG, "===> SHUTDOWN EVT\n");
-				/* Handle this like a regular disconnect */
-				return 0;
-				break;
-			}
-			return -EAGAIN;
-		}
-		msgb_sctp_ppid(msg) = ntohl(sinfo.sinfo_ppid);
-		msgb_sctp_stream(msg) = sinfo.sinfo_stream;
-		break;
-#endif
-	case IPPROTO_TCP:
-	default:
+	switch (conn->srv->sk_domain) {
+	case AF_UNIX:
 		ret = recv(conn->ofd.fd, msgb_data(msg), msgb_tailroom(msg), 0);
 		break;
+	case AF_INET:
+	case AF_INET6:
+	case AF_UNSPEC:
+		switch (conn->srv->proto) {
+#ifdef HAVE_LIBSCTP
+		case IPPROTO_SCTP:
+			ret = _sctp_recvmsg_wrapper(conn->ofd.fd, msg);
+			break;
+#endif
+		case IPPROTO_TCP:
+		default:
+			ret = recv(conn->ofd.fd, msgb_data(msg), msgb_tailroom(msg), 0);
+			break;
+		}
+		break;
+	default:
+		ret = -ENOTSUP;
 	}
 
 	if (ret < 0) {
 		if (errno == EPIPE || errno == ECONNRESET) {
 			LOGP(DLINP, LOGL_ERROR,
-				"lost connection with srv\n");
+				"lost connection with client\n");
 		}
 		return ret;
 	} else if (ret == 0) {
-		LOGP(DLINP, LOGL_ERROR, "connection closed with srv\n");
+		LOGP(DLINP, LOGL_ERROR, "connection closed with client\n");
 		return ret;
 	}
 	msgb_put(msg, ret);
@@ -1371,4 +1624,12 @@ int osmo_stream_srv_recv(struct osmo_stream_srv *conn, struct msgb *msg)
 	return ret;
 }
 
+void osmo_stream_srv_clear_tx_queue(struct osmo_stream_srv *conn)
+{
+	msgb_queue_free(&conn->tx_queue);
+	osmo_fd_write_disable(&conn->ofd);
+	if (conn->flags & OSMO_STREAM_SRV_F_FLUSH_DESTROY)
+		osmo_stream_srv_destroy(conn);
+}
+
 /*! @} */
